# -*- coding: utf-8 -*-
# pylint: disable=all
"""
Least Cost Xmission Command Line Interface

TODO - add cmd line doc

"""
import os
import click
import logging
import warnings
import pandas as pd
import geopandas as gpd
from pathlib import Path

from rex.utilities.loggers import init_mult, create_dirs, init_logger
from rex.utilities.cli_dtypes import STR, INTLIST, INT, FLOAT
from rex.utilities.hpc import SLURM
from rex.utilities.utilities import get_class_properties

from reV.supply_curve.extent import SupplyCurveExtent
from reVX import __version__
from reVX.config.least_cost_xmission import LeastCostXmissionConfig
from reVX.least_cost_xmission.least_cost_xmission import (LeastCostXmission,
                                                          ReinforcedXmission)
from reVX.least_cost_xmission.config import (TRANS_LINE_CAT, LOAD_CENTER_CAT,
                                             SINK_CAT, SUBSTATION_CAT)

TRANS_CAT_TYPES = [TRANS_LINE_CAT, LOAD_CENTER_CAT, SINK_CAT, SUBSTATION_CAT]

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('--name', '-n', default='LeastCostXmission', type=STR,
              show_default=True,
              help='Job name.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, verbose):
    """
    Least Cost Xmission Command Line Interface
    """
    ctx.ensure_object(dict)
    ctx.obj['VERBOSE'] = verbose
    ctx.obj['NAME'] = name


@main.command()
def valid_config_keys():
    """
    Echo the valid Least Cost Xmission config keys
    """
    click.echo(', '.join(get_class_properties(LeastCostXmissionConfig)))


@main.command()
@click.option('--config', '-c', required=True,
              type=click.Path(exists=True),
              help='Filepath to AssemblyAreas config json file.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config, verbose):
    """
    Run Least Cost Xmission from a config.
    """

    config = LeastCostXmissionConfig(config)
    option = config.execution_control.option

    if 'VERBOSE' in ctx.obj:
        if any((ctx.obj['VERBOSE'], verbose)):
            config._log_level = logging.DEBUG
    elif verbose:
        config._log_level = logging.DEBUG

    if option == 'local':
        run_local(ctx, config)
        return

    if option != 'eagle':
        click.echo('Option "{}" is not supported'.format(option))
        return

    if config.execution_control.nodes == 1:
        eagle(config)
        return

    # Split gids over mulitple SLURM jobs
    name = config.name
    logger.info('Splitting SC points over {} SLURM jobs'
                .format(config.execution_control.nodes))
    for i in range(config.execution_control.nodes):
        config.name = '{}_{}'.format(name, i)
        eagle(config, start_index=i)


@main.command()
@click.option('--cost_fpath', '-cost', type=click.Path(exists=True),
              required=True,
              help=("Path to h5 file with cost rasters and other required "
                    "layers"))
@click.option('--features_fpath', '-feats', required=True,
              type=click.Path(exists=True),
              help="Path to GeoPackage with transmission features")
@click.option('--balancing_areas_fpath', '-ba', type=STR, show_default=True,
              default=None,
              help=("Path to Balancing areas GeoPackage. If no `None`, "
                    "Least Cost Xmission is run with reinforcement path "
                    "costs. Features must be substations only, and the "
                    "substation file must contain a 'ba_str' column that "
                    "matches the BA ID in this file for the balancing area "
                    "containing that substation. "))
@click.option('--capacity_class', '-cap', type=str, required=True,
              help=("Capacity class of transmission features to connect "
                    "supply curve points to"))
@click.option('--resolution', '-res', type=int,
              show_default=True, default=128,
              help=("SC point resolution"))
@click.option('--xmission_config', '-xcfg', type=STR, show_default=True,
              default=None,
              help=("Path to Xmission config .json"))
@click.option('--min_line_length', '-mll', type=int,
              show_default=True, default=0,
              help=("Minimum Tie-line length."))
@click.option('--sc_point_start_index', '-start', type=int,
              show_default=True, default=0,
              help=("Start index of supply curve points to run."))
@click.option('--sc_point_step_index', '-step', type=int,
              show_default=True, default=1,
              help=("Step index of supply curve points to run."))
@click.option('--nn_sinks', '-nn', type=int,
              show_default=True, default=2,
              help=("Number of nearest neighbor sinks to use for clipping "
                    "radius calculation. This is overridden by --radius"))
@click.option('--clipping_buffer', '-buffer', type=float,
              show_default=True, default=1.05,
              help=("Buffer to expand clipping radius by"))
@click.option('--barrier_mult', '-bmult', type=float,
              show_default=True, default=100,
              help=("Tranmission barrier multiplier, used when computing the "
                    "least cost tie-line path"))
@click.option('--max_workers', '-mw', type=INT,
              show_default=True, default=None,
              help=("Number of workers to use for processing, if 1 run in "
                    "serial, if None use all available cores"))
@click.option('--out_dir', '-o', type=STR, default='./',
              show_default=True,
              help='Directory to save least cost xmission values to.')
@click.option('--log_dir', '-log', default=None, type=STR,
              show_default=True,
              help='Directory to dump log files.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.option('--save_paths', is_flag=True,
              help='Save least cost paths and data to GeoPackage.')
@click.option('--radius', '-rad', type=INT,
              show_default=True, default=None,
              help=("Radius to clip costs raster to in pixels This overrides "
                    "--nn_sinks if set."))
@click.option('--simplify-geo', type=FLOAT,
              show_default=True, default=None,
              help=("Simplify path geometries by a value before writing to "
                    "GeoPackage."))
@click.pass_context
def local(ctx, cost_fpath, features_fpath, balancing_areas_fpath,
          capacity_class, resolution, xmission_config, min_line_length,
          sc_point_start_index, sc_point_step_index, nn_sinks,
          clipping_buffer, barrier_mult, max_workers, out_dir, log_dir,
          verbose, save_paths, radius, simplify_geo):
    """
    Run Least Cost Xmission on local hardware
    """
    name = ctx.obj['NAME']
    if 'VERBOSE' in ctx.obj:
        verbose = any((ctx.obj['VERBOSE'], verbose))

    log_modules = [__name__, 'reVX', 'reV', 'rex']
    init_mult(name, log_dir, modules=log_modules, verbose=verbose)

    create_dirs(out_dir)
    logger.info('Computing Least Cost Xmission connections and writing them {}'
                .format(out_dir))
    sce = SupplyCurveExtent(cost_fpath, resolution=resolution)
    sc_point_gids = list(sce.points.index.values)
    sc_point_gids = sc_point_gids[sc_point_start_index::sc_point_step_index]
    if balancing_areas_fpath is not None:
        least_costs = ReinforcedXmission.run(cost_fpath, features_fpath,
                                             balancing_areas_fpath,
                                             capacity_class,
                                             resolution=resolution,
                                             xmission_config=xmission_config,
                                             min_line_length=min_line_length,
                                             sc_point_gids=sc_point_gids,
                                             clipping_buffer=clipping_buffer,
                                             barrier_mult=barrier_mult,
                                             max_workers=max_workers,
                                             save_paths=save_paths,
                                             simplify_geo=simplify_geo)
    else:
        least_costs = LeastCostXmission.run(cost_fpath, features_fpath,
                                            capacity_class,
                                            resolution=resolution,
                                            xmission_config=xmission_config,
                                            min_line_length=min_line_length,
                                            sc_point_gids=sc_point_gids,
                                            nn_sinks=nn_sinks,
                                            clipping_buffer=clipping_buffer,
                                            barrier_mult=barrier_mult,
                                            max_workers=max_workers,
                                            save_paths=save_paths,
                                            radius=radius,
                                            simplify_geo=simplify_geo)
    if len(least_costs) == 0:
        logger.error('No paths found.')
        return

    ext = 'gpkg' if save_paths else 'csv'
    fn_out = '{}_{}_{}.{}'.format(name, capacity_class, resolution, ext)
    fpath_out = os.path.join(out_dir, fn_out)

    logger.info('Writing output to {}'.format(fpath_out))
    if save_paths:
        least_costs.to_file(fpath_out, driver='GPKG')
    else:
        least_costs.to_csv(fpath_out, index=False)
    logger.info('Writing output complete')


@main.command()
@click.option('--split-to-geojson', '-s', is_flag=True,
              help='After merging GeoPackages, split into GeoJSON by POI name'
              '.')
@click.option('--out-file', '-of', default=None, type=STR,
              help='Name for output GeoPackage file.')
@click.option('--drop', '-d', default=None, type=STR, multiple=True,
              help=('Transmission feature category types to drop from '
                    'results. Options: {}'.format(", ".join(TRANS_CAT_TYPES))))
@click.option('--out-dir', '-od', type=click.Path(exists=True),
              default='.', show_default=True,
              help='Output directory for output files. Path must exist.')
@click.option('--simplify-geo', type=FLOAT,
              show_default=True, default=None,
              help='Simplify path geometries by a value before exporting.')
@click.argument('files', type=STR, nargs=-1)
@click.pass_context
def merge_output(ctx, split_to_geojson, out_file, out_dir, drop,  # noqa
                 simplify_geo, files):
    """
    Merge output GeoPackage/CSV files and optionally convert to GeoJSON
    """
    log_level = "DEBUG" if ctx.obj.get('VERBOSE') else "INFO"
    init_logger('reVX', log_level=log_level)

    if len(files) == 0:
        logger.info('No files passed to be merged')
        return

    if len(files) == 1:
        files = sorted(Path(out_dir).glob(files[0]))

    logger.debug('Merging {}'.format(files))

    if drop:
        for cat in drop:
            if cat not in TRANS_CAT_TYPES:
                logger.info('--drop options must on or more of {}, received {}'
                            .format(TRANS_CAT_TYPES, drop))
                return

    logger.info('Loading: {}'.format(", ".join(files)))
    warnings.filterwarnings('ignore', category=RuntimeWarning)
    df = pd.concat([gpd.read_file(f) if "gpkg" in f else pd.read_csv(f)
                    for f in files])
    warnings.filterwarnings('default', category=RuntimeWarning)

    if drop:
        mask = df['category'].isin(drop)
        logger.info('Dropping {} of {} total features with category(ies): {}'
                    .format(mask.sum(), len(df), ", ".join(drop)))
        df = df[~mask]

    df = df.reset_index()

    if len(df) == 0:
        logger.info('No transmission features to save.')
        return

    if simplify_geo:
        logger.info('Simplifying geometries by {}'.format(simplify_geo))
        df.geometry = df.geometry.simplify(simplify_geo)

    if not split_to_geojson:
        out_file = ('combo_{}'.format(files[0])
                    if out_file is None else out_file)
        out_file = os.path.join(out_dir, out_file)
        logger.info('Saving to {}'.format(out_file))
        if "gpkg" in out_file:
            df.to_file(out_file, driver="GPKG")
        else:
            df.to_csv(out_file, index=False)
        return

    # Split out put in to GeoJSON by POI name
    for poi in set(df['POI Name']):
        out_file = os.path.join(out_dir,
                                "{}_paths.geojson"
                                .format(poi.replace(' ', '_')))
        paths = df[df['POI Name'] == poi].to_crs(epsg=4326)
        logger.info('Writing {} paths for {} to {}'
                    .format(len(paths), poi, out_file))
        paths.to_file(out_file, driver="GeoJSON")


@main.command()
@click.option('--cost_fpath', '-f', required=True,
              type=click.Path(exists=True),
              help=("Path to GeoPackage/CSV file with calculated transmission "
                    "costs. This file must have a 'trans_gid' column that "
                    "will be used to merge in the reinforcement costs."))
@click.option('--reinforcement_cost_fpath', '-r', required=True,
              type=click.Path(exists=True),
              help=("Path to GeoPackage/CSV file with calculated "
                    "reinforcement costs. This file must have a 'gid' column "
                    "that will be used to merge in the reinforcement costs."))
@click.option('--out_file', '-of', default=None, type=STR,
              help='Name for output GeoPackage/CSV file.')
@click.pass_context
def merge_reinforcement_costs(ctx, cost_fpath, reinforcement_cost_fpath,
                              out_file):
    """
    Merge reinforcement costs into transmission costs.
    """
    log_level = "DEBUG" if ctx.obj.get('VERBOSE') else "INFO"
    init_logger('reVX', log_level=log_level)

    costs = (gpd.read_file(cost_fpath)
             if "gpkg" in cost_fpath
             else pd.read_csv(cost_fpath))
    r_costs = (gpd.read_file(reinforcement_cost_fpath)
               if "gpkg" in reinforcement_cost_fpath
               else pd.read_csv(reinforcement_cost_fpath))

    r_costs.index = r_costs.gid

    logger.info("Merging reinforcement costs into transmission costs...")

    r_cols = ["reinforcement_dist_km", "reinforcement_cost_per_mw"]
    costs[r_cols] = r_costs.loc[costs["trans_gid"], r_cols].values

    logger.info("Writing output to {!r}".format(out_file))

    if "gpkg" in out_file:
        costs.to_file(out_file, driver="GPKG", index=False)
    else:
        costs.to_csv(out_file, index=False)


def get_node_cmd(config, start_index=0):
    """
    Get the node CLI call for Least Cost Xmission

    Parameters
    ----------
    config : reVX.config.least_cost_xmission.LeastCostXmissionConfig
        Least Cost Xmission config object.

    Returns
    -------
    cmd : str
        CLI call to submit to SLURM execution.
    """

    args = ['-n {}'.format(SLURM.s(config.name)),
            'local',
            '-cost {}'.format(SLURM.s(config.cost_fpath)),
            '-feats {}'.format(SLURM.s(config.features_fpath)),
            '-ba {}'.format(SLURM.s(config.balancing_areas_fpath)),
            '-cap {}'.format(SLURM.s(config.capacity_class)),
            '-res {}'.format(SLURM.s(config.resolution)),
            '-xcfg {}'.format(SLURM.s(config.xmission_config)),
            '-mll {}'.format(SLURM.s(config.min_line_length)),
            '-start {}'.format(SLURM.s(start_index)),
            '-step {}'.format(SLURM.s(config.execution_control.nodes or 1)),
            '-nn {}'.format(SLURM.s(config.nn_sinks)),
            '-buffer {}'.format(SLURM.s(config.clipping_buffer)),
            '-bmult {}'.format(SLURM.s(config.barrier_mult)),
            '-mw {}'.format(SLURM.s(config.execution_control.max_workers)),
            '-o {}'.format(SLURM.s(config.dirout)),
            '-log {}'.format(SLURM.s(config.log_directory)),
            ]

    if config.save_paths:
        args.append('--save_paths')
    if config.radius:
        args.append('-rad {}'.format(config.radius))
    if config.simplify_geo:
        args.append('--simplify-geo {}'.format(config.simplify_geo))

    if config.log_level == logging.DEBUG:
        args.append('-v')

    cmd = ('python -m reVX.least_cost_xmission.least_cost_xmission_cli {}'
           .format(' '.join(args)))
    logger.debug('Submitting the following cli call:\n\t{}'.format(cmd))

    return cmd


def run_local(ctx, config):
    """
    Run Least Cost Xmission locally using config

    Parameters
    ----------
    ctx : click.ctx
        click ctx object
    config : reVX.config.least_cost_xmission.LeastCostXmissionConfig
        Least Cost Xmission config object.
    """
    ctx.obj['NAME'] = config.name
    ctx.invoke(local,
               cost_fpath=config.cost_fpath,
               features_fpath=config.features_fpath,
               balancing_areas_fpath=config.balancing_areas_fpath,
               capacity_class=config.capacity_class,
               resolution=config.resolution,
               xmission_config=config.xmission_config,
               min_line_length=config.min_line_length,
               sc_point_start_index=0,
               sc_point_step_index=1,
               nn_sinks=config.nn_sinks,
               clipping_buffer=config.clipping_buffer,
               barrier_mult=config.barrier_mult,
               max_workers=config.execution_control.max_workers,
               out_dir=config.dirout,
               log_dir=config.log_directory,
               verbose=config.log_level,
               radius=config.radius,
               save_paths=config.save_paths,
               simplify_geo=config.simplify_geo,
               )


def eagle(config, start_index=0):
    """
    Run Least Cost Xmission on Eagle HPC.

    Parameters
    ----------
    config : reVX.config.least_cost_xmission.LeastCostXmissionConfig
        Least Cost Xmission config object.
    """
    init_logger('rex', log_level='DEBUG')

    cmd = get_node_cmd(config, start_index)
    name = config.name
    log_dir = config.log_directory
    stdout_path = os.path.join(log_dir, 'stdout/')

    slurm_manager = SLURM()

    logger.info('Running Least Cost Xmission on Eagle with '
                'node name "{}"'.format(name))
    out = slurm_manager.sbatch(cmd,
                               alloc=config.execution_control.allocation,
                               memory=config.execution_control.memory,
                               walltime=config.execution_control.walltime,
                               feature=config.execution_control.feature,
                               name=name, stdout_path=stdout_path,
                               conda_env=config.execution_control.conda_env,
                               module=config.execution_control.module)[0]
    if out:
        msg = ('Kicked off Least Cost Xmission "{}" '
               '(SLURM jobid #{}) on Eagle.'
               .format(name, out))
    else:
        msg = ('Was unable to kick off Least Cost Xmission '
               '"{}". Please see the stdout error messages'
               .format(name))

    click.echo(msg)
    logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running Least Cost Xmission CLI')
        raise



# -*- coding: utf-8 -*-
# pylint: disable=too-many-arguments,too-many-locals
"""
Turbine Flicker exclusions calculator
"""
from concurrent.futures import as_completed
import logging
import numpy as np
import os
from warnings import warn

from reV.handlers.exclusions import ExclusionLayers
from reV.supply_curve.extent import SupplyCurveExtent
from reV.supply_curve.tech_mapping import TechMapping
from reVX.handlers.geotiff import Geotiff
from reVX.wind_dirs.mean_wind_dirs_point import MeanWindDirectionsPoint
from reVX.utilities.exclusions import AbstractBaseExclusionsMerger
from rex.resource_extraction.resource_extraction import WindX
from rex.utilities.execution import SpawnProcessPool
from rex.utilities.loggers import log_mem

logger = logging.getLogger(__name__)


class TurbineFlicker(AbstractBaseExclusionsMerger):
    """
    Class to compute turbine shadow flicker and exclude sites that will
    cause excessive flicker on building
    """
    STEPS_PER_HOUR = 1
    DEFAULT_FEATURE_OUTFILE = 'flicker.tif'

    def __init__(self, excl_fpath, res_fpath, features, regulations,
                 building_threshold=0, resolution=640, grid_cell_size=90,
                 max_flicker_exclusion_range="10x", tm_dset='techmap_wtk',
                 hsds=False):
        """
        Parameters
        ----------
        excl_fpath : str
            Filepath to exclusions h5 file. File must contain
            "building_layer" and "tm_dset".
        res_fpath : str
            Filepath to wind resource .h5 file containing hourly wind
            direction data
        features : str
            This input should either be the name of an exclusion layer
            in `excl_fpath` or a file path to a GeoTIFF file containing
            buildings data from which turbine flicker exclusions should
            be computed.
        regulations : `FlickerRegulations`
            A `FlickerRegulations` object used to shadow flicker
            regulation values.
        building_threshold : float, optional
            Threshold for exclusion layer values to identify pixels with
            buildings, values are % of pixel containing a building.
            Threshold is NOT inclusive. By default, `0`.
        resolution : int, optional
            SC resolution, must be input in combination with gid,
            by default 640
        grid_cell_size : float, optional
            Length (m) of a side of each grid cell in `excl_fpath`.
        max_flicker_exclusion_range : float | int | str, optional
            Max distance (m) that flicker exclusions will extend in
            any of the cardinal directions. Can also be a string like
            ``"10x"`` (default), which is interpreted as 10 times the
            turbine rotor diameter. Note that increasing this value can
            lead to drastically instead memory requirements. This value
            may be increased slightly during execution (no more then the
            size of one grid cell) in order to yield odd exclusion array
            shapes.
        tm_dset : str, optional
            Dataset / layer name for wind toolkit techmap,
            by default 'techmap_wtk'
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on
            AWS behind HSDS. By default `False`.
        """
        super().__init__(excl_fpath, regulations, features=features,
                         hsds=hsds)
        self._res_h5 = res_fpath
        self._res = resolution
        self._grid_cell_size = grid_cell_size
        self._max_flicker_exclusion_range = (
            self._parse_max_flicker_exclusion_range(
                max_flicker_exclusion_range))
        self._set_max_grid_size_for_odd_shaped_arr()
        self._flicker_preflight_check(tm_dset=tm_dset)
        self._sc_points = self._get_sc_points(tm_dset=tm_dset)
        self._fips_to_gid = {}
        self.features = self.features > building_threshold

    def _parse_max_flicker_exclusion_range(self, excl_range):
        """Convert max_flicker_exclusion_range to float if necessary. """
        if isinstance(excl_range, str) and excl_range.endswith('x'):
            rd = self._regulations.rotor_diameter
            return float(excl_range.strip('x')) * rd

        if not isinstance(excl_range, (int, float)):
            try:
                excl_range = float(excl_range)
            except Exception as e:
                msg = ('max_flicker_exclusion_range must be numeric but '
                       'received: {}, {}'.format(excl_range, type(excl_range)))
                logger.error(msg)
                raise TypeError(msg) from e

        return excl_range

    def _flicker_preflight_check(self, tm_dset='techmap_wtk'):
        """
        Check to ensure building_layer and tm_dset are in exclusion .h5 file

        Parameters
        ----------
        tm_dset : str, optional
            Dataset / layer name for wind toolkit techmap,
            by default 'techmap_wtk'
        """
        with ExclusionLayers(self._excl_fpath, hsds=self._hsds) as f:
            layers = f.layers
            exclusion_shape = f.shape

        if self.features.shape != exclusion_shape:
            msg = ("Shape of building layer {} does not match shape of "
                   "ExclusionLayers {}"
                   .format(self.features.shape, exclusion_shape))
            logger.error(msg)
            raise RuntimeError(msg)

        if tm_dset not in layers:
            logger.warning('Could not find techmap "{t}" in {e}. '
                           'Creating {t} using reV TechMapping'
                           .format(t=tm_dset, e=self._excl_fpath))
            try:
                TechMapping.run(self._excl_fpath, self._res_h5,
                                dset=tm_dset)
            except Exception as e:
                logger.exception('TechMapping process failed. Received the '
                                 'following error:\n{}'.format(e))
                raise e

    def _get_sc_points(self, tm_dset='techmap_wtk'):
        """
        Get the valid sc points to run turbine flicker for

        Parameters
        ----------
        tm_dset : str, optional
            Dataset / layer name for wind toolkit techmap, by default
            'techmap_wtk'

        Returns
        -------
        points : pandas.DataFrame
            DataFrame of valid sc point gids with their latitude and longitude
            coordinates and nearest resource gid
        """
        with SupplyCurveExtent(self._excl_fpath, resolution=self._res) as sc:
            points = sc.points
            points['latitude'] = sc.latitude
            points['longitude'] = sc.longitude
            gids = sc.valid_sc_points(tm_dset)
            points = points.loc[gids]

        with WindX(self._res_h5) as f:
            res_gids = f.lat_lon_gid(points[['latitude', 'longitude']].values,
                                     check_lat_lon=False)
            points['res_gid'] = res_gids

        return points

    def _set_max_grid_size_for_odd_shaped_arr(self):
        """Set the max_flicker_exclusion_range to multiple of 0.5 grids """
        mult = np.round(self._max_flicker_exclusion_range
                        / self._grid_cell_size) + 0.5
        self._max_flicker_exclusion_range = mult * self._grid_cell_size

    @property
    def description(self):
        """str: Description to be added to excl H5."""
        return ('Pixels with value 0 are excluded as they will cause shadow '
                'flicker on buildings. Shadow flicker is computed using a '
                '{}m hub height, {}m rotor diameter turbine.'
                .format(self._regulations.hub_height,
                        self._regulations.rotor_diameter))

    def _apply_regulations_mask(self):
        """Mask regulations to only shadow flicker. """
        flicker = self._regulations.df['Feature Type'] == 'shadow flicker'

        if not flicker.any():
            msg = "Found no local flicker regulations!"
            logger.warning(msg)
            warn(msg)

        self._regulations.df = (self._regulations.df[flicker]
                                .reset_index(drop=True))
        logger.debug('Computing flicker for regulations in {} counties'
                     .format(len(self._regulations.df)))

    def _map_fips_to_gid(self):
        """Map county FIPS values to corresponding SC gids. """

        self._fips_to_gid = {}
        reg_fips = self._regulations.df.FIPS.unique()
        with SupplyCurveExtent(self._excl_fpath, resolution=self._res) as sc:
            for gid in self._sc_points.index:
                for fips in np.unique(sc.get_excl_points('cnty_fips', gid)):
                    if fips in reg_fips:
                        self._fips_to_gid.setdefault(fips, []).append(gid)

        missing_fips = set(reg_fips) - set(self._fips_to_gid)
        if missing_fips:
            msg = ("{} counties with flicker regulations were not found on "
                   "the supply curve grid ({}): {}"
                   .format(len(missing_fips), self._excl_fpath, missing_fips))
            logger.warning(msg)
            warn(msg)

    def _points(self, fips=None):
        """Points for which to compute shadow flicker. """
        if fips is None:
            gids = self._sc_points.index
        else:
            gids = self._fips_to_gid.get(fips, [])
        return [self._sc_points.loc[gid] for gid in gids]

    @property
    def no_exclusions_array(self):
        """np.array: Array representing no exclusions. """
        return np.ones(self.features.shape, dtype=np.uint8)

    @property
    def exclusion_merge_func(self):
        """callable: Function to merge overlapping exclusion layers. """
        return np.minimum

    def pre_process_regulations(self):
        """Reduce regulations to correct state and features. """
        self._apply_regulations_mask()
        self._map_fips_to_gid()

    def parse_features(self):
        """Get the building layer used to compute flicker exclusions. """
        if os.path.exists(self._features):
            logger.debug("Loading building data from {}"
                         .format(self._features))
            with Geotiff(self._features) as f:
                return f.values[0]

        with ExclusionLayers(self._excl_fpath, hsds=self._hsds) as f:
            if self._features not in f.layers:
                msg = ("{} is not available in {}"
                       .format(self._features, self._excl_fpath))
                logger.error(msg)
                raise RuntimeError(msg)
            logger.debug("Loading building data from {}, layer {}"
                         .format(self._excl_fpath, self._features))
            return f[self._features]

    def _local_exclusions_arguments(self, regulation_value, cnty):
        """Compile and return arguments to `compute_local_exclusions`.

        This method should return a list or tuple of extra args to be
        passed to `compute_local_exclusions`.

        Parameters
        ----------
        regulation_value : float | int
            Regulation value for county.
        cnty : geopandas.GeoDataFrame
            Regulations for a single county.

        Returns
        -------
        TurbineFlicker
            `TurbineFlicker` object used as an extra argument to
            calculate local flicker exclusions.
        """
        return (self._regulations.hub_height,
                self._regulations.rotor_diameter,
                self._points(cnty.iloc[0]['FIPS']),
                self._res_h5,
                self._max_flicker_exclusion_range,
                self._grid_cell_size,
                self.STEPS_PER_HOUR,
                self.features,
                self._res)

    @staticmethod
    def compute_local_exclusions(regulation_value, cnty, *args):
        """Compute local flicker exclusions.

        This method computes a flicker exclusion layer using the
        information about the input county.

        Parameters
        ----------
        regulation_value : float | int
            Maximum number of allowable flicker hours in county.
        cnty : geopandas.GeoDataFrame
            Regulations for a single county.
        tf : `TurbineFlicker`
            Instance of `TurbineFlicker` objects used to compute the
            flicker exclusions.

        Returns
        -------
        flicker : ndarray
            Raster array of flicker exclusions
        """
        (hub_height, rotor_diameter, points, res_fpath,
         max_flicker_exclusion_range, grid_cell_size, steps_per_hour,
         building_layer, resolution) = args
        logger.debug('- Computing flicker for county FIPS {}'
                     .format(cnty.iloc[0]['FIPS']))
        return compute_flicker_exclusions(hub_height, rotor_diameter, points,
                                          res_fpath, regulation_value,
                                          max_flicker_exclusion_range,
                                          grid_cell_size, steps_per_hour,
                                          building_layer, resolution,
                                          max_workers=1)

    def compute_generic_exclusions(self, max_workers=None):
        """Compute generic flicker exclusions.

        This method will compute a generic flicker exclusion layer.

        Parameters
        ----------
        max_workers : int, optional
            Number of workers to use for exclusions computation, if 1
            run in serial, if > 1 run in parallel with that many
            workers, if `None` run in parallel on all available cores.
            By default `None`.

        Returns
        -------
        flicker : ndarray
            Raster array of flicker exclusions
        """
        logger.info('Computing generic flicker exclusions using a threshold '
                    'of from {} hrs/year based on {}m hub height, {}m '
                    'rotor diameter turbines'
                    .format(self._regulations.generic,
                            self._regulations.hub_height,
                            self._regulations.rotor_diameter))
        return compute_flicker_exclusions(self._regulations.hub_height,
                                          self._regulations.rotor_diameter,
                                          self._points(), self._res_h5,
                                          self._regulations.generic,
                                          self._max_flicker_exclusion_range,
                                          self._grid_cell_size,
                                          self.STEPS_PER_HOUR,
                                          self.features, self._res,
                                          max_workers=max_workers)

    @classmethod
    def input_output_filenames(cls, out_dir, features_fpath, kwargs):
        """Generate pairs of input/output file names.

        Parameters
        ----------
        out_dir : str
            Path to output file directory.
        features_fpath : str
            This input should either be the name of an exclusion layer
            in `excl_fpath` or a file path to a GeoTIFF file containing
            buildings data from which turbine flicker exclusions should
            be computed.
        kwargs : dict
            Dictionary of extra keyword-argument pairs used to
            instantiate the `exclusion_class`.

        Yields
        ------
        tuple
            An input-output filename pair.
        """
        regulations = kwargs['regulations']
        fn = flicker_fn_out(regulations.hub_height, regulations.rotor_diameter)
        yield features_fpath, os.path.join(out_dir, fn)


def _get_building_indices(building_layer, gid, resolution=640):
    """Find buildings exclusion indices

    Parameters
    ----------
    building_layer : np.ndarray
        Exclusion layer containing buildings from which turbine flicker
        exclusions will be computed.
    gid : int
        SC point gid to extract buildings for.
    resolution : int, optional
        SC resolution, must be input in combination with gid.
        By default, `640`.

    Returns
    -------
    row_idx : ndarray
        Axis 0 indices of building in sc point sub-array in full
        exclusion array.
    col_idx : ndarray
        Axis 1 indices of building in sc point sub-array in full
        exclusion array.
    shape : tuple
        Full exclusion array shape.
    """
    row_slice, col_slice = MeanWindDirectionsPoint.get_agg_slices(
        gid, building_layer.shape, resolution
    )
    sc_blds = building_layer[row_slice, col_slice]
    row_idx = np.array(range(*row_slice.indices(row_slice.stop)))
    col_idx = np.array(range(*col_slice.indices(col_slice.stop)))

    bld_row_idx, bld_col_idx = np.where(sc_blds)
    return row_idx[bld_row_idx], col_idx[bld_col_idx]


def _create_excl_indices(bld_idx, flicker_shifts, shape):
    """
    Create 2D (row, col) indices of pixels to be excluded based on
    building indices and shadow flicker shifts.

    Parameters
    ----------
    bld_idx : tuple
        (row, col) indices of building onto which shadow flicker
        exclusions are to be mapped.
    flicker_shifts : tuple
        Index shifts (row, col) from building locations to exclude based
        on shadow flicker results. Shifts are based on shadow flicker
        threshold. Shadow flicker array is inverted to represent mapping
        of shadow onto buildings
    shape : tuple
        Full exclusion array shape

    Returns
    -------
    excl_row_idx : ndarray
        Row (axis 0) indices of pixels to be excluded because they will
        cause excessive shadow flicker on building in supply curve point
        gid subset
    excl_col_idx : ndarray
        Column (axis 1) indices of pixels to be excluded because they
        will cause excessive shadow flicker on building in supply curve
        point gid subset
    """
    row_idx, col_idx = bld_idx
    row_shifts, col_shifts = flicker_shifts

    excl_row_idx = (row_idx + row_shifts[:, None]).ravel()
    excl_row_idx[excl_row_idx < 0] = 0
    excl_row_idx[excl_row_idx >= shape[0]] = shape[0] - 1

    excl_col_idx = (col_idx + col_shifts[:, None]).ravel()
    excl_col_idx[excl_col_idx < 0] = 0
    excl_col_idx[excl_col_idx >= shape[1]] = shape[1] - 1

    return excl_row_idx.astype(np.int32), excl_col_idx.astype(np.int32)


def _invert_shadow_flicker_arr(shadow_flicker):
    """
    Check to ensure the shadow_flicker array is odd in shape, i.e. both
    dimensions are odd allowing for a central pixel for the turbine to
    sit on. Flip both axes to mimic the turbine sitting on each
    building. All flicker pixels will now indicate locations where a
    turbine would need to be to cause flicker on said building

    Parameters
    ----------
    shadow_flicker : ndarray
        2D array centered on the turbine with the number of flicker
        hours per "exclusion" pixel

    Returns
    -------
    shadow_flicker : ndarray
        Inverted 2D shadow flicker array with odd dimensions if needed.
    """
    reduce_slice = ()
    reduce_arr = False
    for s in shadow_flicker.shape:
        if s % 2:
            reduce_slice += (slice(None), )
        else:
            reduce_slice += (slice(0, -1), )
            reduce_arr = True

    if reduce_arr:
        shape_in = shadow_flicker.shape
        shadow_flicker = shadow_flicker[reduce_slice]
        msg = ('Shadow flicker array with shape {} does not have a '
               'central pixel! Shade has been reduced to {}!'
               .format(shape_in, shadow_flicker.shape))
        logger.warning(msg)
        warn(msg)

    return shadow_flicker[::-1, ::-1]


def _get_flicker_excl_shifts(shadow_flicker, flicker_threshold=30):
    """
    Determine locations of shadow flicker that exceed the given threshold,
    convert to row and column shifts. These are the locations turbines
    would need to in relation to building to cause flicker exceeding the
    threshold value.

    Parameters
    ----------
    shadow_flicker : ndarray
        2D array centered on the turbine with the number of flicker hours
        per "exclusion" pixel
    flicker_threshold : int, optional
        Maximum number of allowable flicker hours, by default 30

    Returns
    -------
    row_shifts : ndarray
        Shifts along axis 0 from building location to pixels to be excluded
    col_shifts : ndarray
        Shifts along axis 1 from building location to pixels to be excluded
    """
    # ensure shadow_flicker array is regularly shaped and invert for
    # mapping to buildings
    shadow_flicker = _invert_shadow_flicker_arr(shadow_flicker)

    # normalize by number of time-steps to match shadow flicker results
    flicker_threshold /= 8760
    shape = shadow_flicker.shape
    row_shifts, col_shifts = np.where(shadow_flicker > flicker_threshold)
    check = (np.any(np.isin(row_shifts, [0, shape[0] - 1]))
             or np.any(np.isin(col_shifts, [0, shape[1] - 1])))
    if check:
        msg = ("Turbine flicker appears to extend beyond the FlickerModel "
               "domain! Consider increasing the maximum flicker exclusion "
               "range.")
        logger.warning(msg)
        warn(msg)

    row_shifts -= shape[0] // 2
    col_shifts -= shape[1] // 2

    return row_shifts, col_shifts


def _compute_shadow_flicker(rotor_diameter, lat, lon, wind_dir,
                            max_flicker_exclusion_range, grid_cell_size,
                            steps_per_hour=1):
    """
    Compute shadow flicker for given location

    Parameters
    ----------
    rotor_diameter : int | float
        Rotor diameter in meters.
    lat : float
        Latitude coordinate of turbine.
    lon : float
        Longitude coordinate of turbine.
    wind_dir : ndarray
        Time-series of wind direction for turbine.
    max_flicker_exclusion_range : float | int | str
        Max distance (m) that flicker exclusions will extend in
        any of the cardinal directions.
    grid_cell_size : int
        Length (m) of a side of each grid cell in `excl_fpath`.
    steps_per_hour : int
        Number of time steps to take per hour when computing flicker.

    Returns
    -------
    shadow_flicker : ndarray
        2D array centered on the turbine with the number of flicker
        hours per "exclusion" pixel
    """
    # Import HOPP dynamically so its not a requirement
    from hybrid.flicker.flicker_mismatch_grid import FlickerMismatch

    mult = max_flicker_exclusion_range / rotor_diameter
    FlickerMismatch.diam_mult_nwe = mult
    FlickerMismatch.diam_mult_s = mult
    FlickerMismatch.steps_per_hour = steps_per_hour
    FlickerMismatch.turbine_tower_shadow = False

    assert len(wind_dir) == 8760

    shadow_flicker = FlickerMismatch(lat, lon,
                                     blade_length=rotor_diameter / 2,
                                     angles_per_step=None,
                                     wind_dir=wind_dir,
                                     gridcell_height=grid_cell_size,
                                     gridcell_width=grid_cell_size,
                                     gridcells_per_string=1)
    shadow_flicker = shadow_flicker.create_heat_maps(range(0, 8760),
                                                     ("time", ))[0]

    return shadow_flicker


def _exclude_turbine_flicker(hub_height, rotor_diameter, point, res_fpath,
                             flicker_threshold, max_flicker_exclusion_range,
                             grid_cell_size, steps_per_hour):
    """
    Exclude all pixels that will cause flicker exceeding the
    "flicker_threshold" on buildings that exist within
    supply curve point gid subset of "building_layer". Buildings are
    defined as any pixels in the array that are "truthy". In other
    words, the locations of the buildings are found using
    `np.where(building_layer)`. Shadow flicker is computed at the supply
    curve point resolution and applied to all buildings within that
    supply curve point sub-array. Excluded pixels can extend beyond the
    supply curve point gid subset, for example if a building sits at the
    edge of the subset.

    Parameters
    ----------
    hub_height : int | float
        Hub height in meters.
    rotor_diameter : int | float
        Rotor diameter in meters.
    gid : int
        Supply curve point gid to aggregate wind directions for
    res_fpath : str
        Filepath to wind resource .h5 file containing hourly wind
        direction data
    flicker_threshold : int
        Maximum number of allowable flicker hours.
    max_flicker_exclusion_range : float | int | str
        Max distance (m) that flicker exclusions will extend in
        any of the cardinal directions.
    grid_cell_size : int
        Length (m) of a side of each grid cell in `excl_fpath`.
    steps_per_hour : int
        Number of time steps to take per hour when computing flicker.

    Returns
    -------
    excl_idx : tuple
        (row, col) shifts of pixels to be excluded because they
        will cause excessive shadow flicker from building location
    """

    with WindX(res_fpath, log_vers=False) as f:
        dset = 'winddirection_{}m'.format(hub_height)
        wind_dir = f[dset, :, int(point['res_gid'])]

        # pylint: disable=unsubscriptable-object
        if len(wind_dir) == 8784:
            wind_dir = wind_dir[:-24]

    shadow_flicker = _compute_shadow_flicker(rotor_diameter, point['latitude'],
                                             point['longitude'], wind_dir,
                                             max_flicker_exclusion_range,
                                             grid_cell_size, steps_per_hour)

    thresh = flicker_threshold
    flicker_shifts = _get_flicker_excl_shifts(shadow_flicker,
                                              flicker_threshold=thresh)

    return flicker_shifts


def compute_flicker_exclusions(hub_height, rotor_diameter, points, res_fpath,
                               flicker_threshold, max_flicker_exclusion_range,
                               grid_cell_size, steps_per_hour, building_layer,
                               resolution=640, max_workers=None):
    """Compute turbine flicker exclusions.

    Exclude all pixels that will cause flicker exceeding the
    "flicker_threshold" on any building in "building_layer".
    Buildings are defined as any pixels in the array that are "truthy".
    In other words, the locations of the buildings are found using
    `np.where(building_layer)`. Shadow flicker is computed at the
    supply curve point resolution based on a turbine with
    "hub_height" (m) and applied to all buildings within that supply
    curve point sub-array.

    Parameters
    ----------
    hub_height : int | float
        Hub height in meters.
    rotor_diameter : int | float
        Rotor diameter in meters.
    points : iterable of pd.Series
        Supply curve points to calculate shadow flicker exclusions for.
        The series for each point should include, at the minimum,
        "res_gid", "latitude" and "longitude".
    res_fpath : str
        Filepath to wind resource .h5 file containing hourly wind
        direction data.
    flicker_threshold : int
        Maximum number of allowable flicker hours.
    max_flicker_exclusion_range : float | int | str
        Max distance (m) that flicker exclusions will extend in
        any of the cardinal directions.
    grid_cell_size : int
        Length (m) of a side of each grid cell in `excl_fpath`.
    steps_per_hour : int
        Number of time steps to take per hour when computing flicker.
    building_layer : np.ndarray
        Array containing building data. Any "truthy" values in this
        array are assumed to be buildings, so be sure to apply any
        appropriate threshold/mask before passing in the array.
    resolution : int, optional
        SC resolution, by default 640.
    max_workers : int, optional
        Number of workers to use. If 1 run, in serial. If `None`,
        use all available cores. By default, `None`.

    Returns
    -------
    flicker_arr : ndarray
        2D inclusion array. Pixels to exclude (0) to prevent shadow
        flicker on buildings in "building_layer
    """

    er = max_flicker_exclusion_range
    if max_workers is None:
        max_workers = os.cpu_count()

    flicker_arr = np.ones(building_layer.shape, dtype=np.uint8)
    if max_workers > 1:
        msg = ('Computing flicker exclusions based on {}m hub '
               'height turbines with {}m rotor diameters in parallel '
               'using {} workers'
               .format(hub_height, rotor_diameter, max_workers))
        logger.info(msg)

        loggers = [__name__, 'reVX', 'rex']
        with SpawnProcessPool(max_workers=max_workers, loggers=loggers) as exe:
            futures = {}
            for point in points:
                future = exe.submit(_exclude_turbine_flicker, hub_height,
                                    rotor_diameter, point, res_fpath,
                                    flicker_threshold, er, grid_cell_size,
                                    steps_per_hour)
                futures[future] = point

            for i, future in enumerate(as_completed(futures)):
                flicker_shifts = future.result()
                point = futures.pop(future)
                row_idx, col_idx = _get_building_indices(building_layer,
                                                         point.name,
                                                         resolution=resolution)
                row_idx, col_idx = _create_excl_indices((row_idx, col_idx),
                                                        flicker_shifts,
                                                        building_layer.shape)

                flicker_arr[row_idx, col_idx] = 0
                logger.info('Completed {} out of {} gids'
                            .format((i + 1), len(futures)))
                log_mem(logger)
    else:
        msg = ('Computing local flicker exclusions based on {}m hub height, '
               '{}m rotor diameter turbines in serial.'
               .format(hub_height, rotor_diameter))
        logger.info(msg)
        for i, point in enumerate(points):
            flicker_shifts = _exclude_turbine_flicker(hub_height,
                                                      rotor_diameter, point,
                                                      res_fpath,
                                                      flicker_threshold, er,
                                                      grid_cell_size,
                                                      steps_per_hour)

            row_idx, col_idx = _get_building_indices(building_layer,
                                                     point.name,
                                                     resolution=resolution)
            row_idx, col_idx = _create_excl_indices((row_idx, col_idx),
                                                    flicker_shifts,
                                                    building_layer.shape)

            flicker_arr[row_idx, col_idx] = 0
            logger.debug('Completed {} out of {} gids'
                         .format((i + 1), len(points)))
            log_mem(logger)

    return flicker_arr


def flicker_fn_out(hub_height, rotor_diameter):
    """Generate flicker tiff outfile name.

    Parameters
    ----------
    hub_height : int
        Turbine hub-height (m).
    rotor_diameter : int
        Turbine rotor diameter (m).

    Returns
    -------
    str
        Name of flicker outfile.
    """
    return "flicker_{}hh_{}rd.tif".format(hub_height, rotor_diameter)



"""
Region Classifier Module
"""
import pandas as pd
import geopandas as gpd
from shapely.geometry import Point
from shapely.geometry import shape
from scipy.spatial import cKDTree
import logging

from reVX.utilities.utilities import log_versions
from rex import Resource

logger = logging.getLogger(__name__)


class RegionClassifier():
    """
    Base class of region classification

    Examples
    --------
    >>> meta_path = 'meta.csv'
    >>> regions_path = 'us_states.shp'
    >>> regions_label = 'NAME'
    >>>
    >>> classifier = RegionClassifier(meta_path=meta_path,
                                      regions_path=regions_path,
                                      regions_label=regions_label)
    >>>
    >>> force = True
    >>> fout = 'new_meta.csv'
    >>>
    >>> classification = classifier.classify(force=force)
    >>> classifier.output_to_csv(classification, fout)
    """

    CRS = "EPSG:4326"
    DEFAULT_REGIONS_LABEL = 'regions_index'

    def __init__(self, meta_path, regions, regions_label=None,
                 outlier_value=-999):
        """
        Parameters
        ----------
        meta_path : str | pandas.DataFrame
            Path to meta CSV file, resource .h5 file, or pre-loaded meta
            DataFrame containing lat/lon points
        regions : str | GeoDataFrame
            Path to regions shapefile containing labeled geometries or
            a pre-loaded GeoDataFrame
        regions_label : str
            Attribute to use as label in the regions shapefile
        outlier_value : float | int | str
            Value to assign to outliers if not force
        """
        log_versions(logger)
        self._regions_label = regions_label
        if self._regions_label is None:
            self._regions_label = self.DEFAULT_REGIONS_LABEL

        self._meta = self._get_meta(meta_path)
        self._regions = self._get_regions(regions, self._regions_label)
        self._outlier_value = outlier_value

    @property
    def regions(self):
        """Get the regions GeoDataFrame

        Returns
        -------
        GeoDataFrame
        """
        return self._regions

    @staticmethod
    def output_to_csv(gdf, path):
        """ Export a geopandas dataframe to csv

        Parameters
        ----------
        gdf : GeoPandas DataFrame
            Meta data to export
        path : str
            Output CSV file path for labeled meta CSV file
        """
        output_gdf = gdf.drop('geometry', axis=1)
        if output_gdf.index.name == 'gid':
            output_gdf = output_gdf.reset_index()

        output_gdf.to_csv(path, index=False)

    @classmethod
    def _get_regions(cls, regions, regions_label):
        """ Load the regions shapefile into geopandas dataframe

        Parameters
        ----------
        regions : str | GeoDataFrame
            Path to regions shapefile containing labeled geometries or
            a pre-loaded GeoDataFrame
        regions_label : str
            Attribute to use as label in the regions shapefile
        """

        if not isinstance(regions, gpd.GeoDataFrame):
            regions = gpd.read_file(regions).to_crs(cls.CRS)
            if regions_label not in regions.columns:
                regions_label = cls.DEFAULT_REGIONS_LABEL
                regions[regions_label] = regions.index
                logger.warning('Setting regions label: {}'
                               .format(regions_label))

        centroids = regions.geometry.centroid
        regions['longitude'] = centroids.x
        regions['latitude'] = centroids.y

        return regions

    @classmethod
    def _get_meta(cls, meta_path):
        """ Load the meta csv file into geopandas dataframe

        Parameters
        ----------
        meta_path : str | pandas.DataFrame
            Path to meta CSV file, resource .h5 file, or pre-loaded meta
            DataFrame containing lat/lon points
        """
        if isinstance(meta_path, str):
            if meta_path.endswith('.csv'):
                meta = pd.read_csv(meta_path)
            elif meta_path.endswith('.h5'):
                with Resource(meta_path) as f:
                    meta = f.meta
            else:
                msg = ("Cannot parse meta data from {}, expecting a .csv or "
                       ".h5 file!".format(meta_path))
                logger.error(msg)
                raise RuntimeError(msg)
        elif isinstance(meta_path, pd.DataFrame):
            meta = meta_path
        else:
            msg = ("Cannot parse meta data from {}, expecting a .csv or "
                   ".h5 file path, or a pre-loaded pandas DataFrame"
                   .format(meta_path))
            logger.error(msg)
            raise RuntimeError(msg)

        lat_label, long_label = cls._get_lat_lon_labels(meta)
        geometry = [Point(xy) for xy in zip(meta[long_label],
                                            meta[lat_label])]
        meta = gpd.GeoDataFrame(meta, crs=cls.CRS,
                                geometry=geometry)

        return meta

    @staticmethod
    def _get_lat_lon_labels(df):
        """ Auto detect the latitude and longitude columns from DataFrame

        Parameters
        ----------
        df : Pandas DataFrame
            Meta data with the latitude/longitude columns to detect
        """

        # Latitude
        lat_col = [c for c in df if c.lower().startswith('lat')]
        if len(lat_col) > 1:
            msg = "Multiple latitude columns found: {}".format(lat_col)
            logger.error(msg)
            raise RuntimeError(msg)

        lat_col = lat_col[0]

        # Longitude
        lon_col = [c for c in df if c.lower().startswith('lon')]
        if len(lon_col) > 1:
            msg = "Multiple longitude columns found: {}".format(lon_col)
            logger.error(msg)
            raise RuntimeError(msg)

        lon_col = lon_col[0]

        return [lat_col, lon_col]

    @staticmethod
    def _nearest(target, lookup):
        """ Lookup the indices to the nearest point

        Parameters
        ----------
        target : Pandas DataFrame
            List of lat/lon points
        lookup : Pandas DataFrame
            List of lat/lon points
        """
        tree = cKDTree(lookup)  # pylint: disable=not-callable
        _, inds = tree.query(target, k=1)

        return inds

    @staticmethod
    def _geom_is_valid(geom):
        """ Check if individual geometry is valid """
        try:
            shape(geom)
            return 1
        except AttributeError:
            return 0

    def classify(self, force=False):
        """ Classify the meta data with regions labels

        Parameters
        ----------
        force : str
            Force outlier classification by finding nearest
        """
        try:
            # Get intersection classifications
            meta_inds, region_inds, outlier_inds = self._intersect()
        except Exception as e:
            logger.warning(e)
            invalid_geom_ids = self._check_geometry()
            if invalid_geom_ids:
                logger.warning('The following geometries are invalid: {}'
                               .format(invalid_geom_ids))
            else:
                logger.exception('Cannot run region classification')
                raise

        classified_meta = self._meta.copy()
        classified_meta[self._regions_label] = self._outlier_value
        region_labels = self._regions.loc[region_inds, self._regions_label]
        region_labels = list(region_labels)
        classified_meta.loc[meta_inds, self._regions_label] = region_labels

        # Check for any intersection outliers
        num_outliers = len(outlier_inds)
        if (num_outliers and force):
            # Lookup the nearest region geometry (by centroid)
            logger.warning('The following points are outliers:')
            logger.warning(outlier_inds)
            cols = self._get_lat_lon_labels(self._meta)
            lookup = self._regions[['latitude', 'longitude']]
            target = self._meta.loc[outlier_inds][cols]
            out_inds = list(self._nearest(target, lookup))

            regions = self._regions.loc[out_inds, self._regions_label]
            regions = list(regions)
            classified_meta.loc[outlier_inds, self._regions_label] = regions

        return classified_meta

    def _intersect(self):
        """ Join the meta points to regions by spatial intersection """

        joined = gpd.sjoin(self._meta, self._regions,
                           how='inner', op='intersects')
        if 'index_left' in joined.columns:
            joined = joined.drop_duplicates('index_left', keep='last')
            meta_inds = list(joined['index_left'])
        else:
            meta_inds = list(joined.index)

        region_inds = list(joined['index_right'])
        outliers = self._meta.loc[~self._meta.index.isin(meta_inds)]
        outlier_inds = list(outliers.index)

        return meta_inds, region_inds, outlier_inds

    def _check_geometry(self):
        """ Get index list of invalid geometries """
        geometry = self._regions.geometry
        isvalid = geometry.apply(self._geom_is_valid)

        return list(self._regions[isvalid == 0].index)

    @classmethod
    def run(cls, meta_path, regions, regions_label=None,
            force=False, fout=None):
        """ Run full classification

        Parameters
        ----------
        meta_path : str | pandas.DataFrame
            Path to meta CSV file, resource .h5 file, or pre-loaded meta
            DataFrame containing lat/lon points
        regions : str | GeoDataFrame
            Path to regions shapefile containing labeled geometries or
            a pre-loaded GeoDataFrame
        regions_label : str
            Attribute to use a label in the regions shapefile
        force : str
            Force outlier classification by finding nearest
        fout : str
            Output CSV file path for labeled meta CSV file

        Examples
        --------
        >>> meta_path = 'meta.csv'
        >>> regions_path = 'us_states.shp'
        >>> regions_label = 'NAME'
        >>> force = True
        >>> fout = 'new_meta.csv'
        >>>
        >>> RegionClassifier.run(meta_path=meta_path,
                                 regions_path=regions_path,
                                 regions_label=regions_label
                                 force=force, fout=fout)
        """
        classifier = cls(meta_path=meta_path, regions=regions,
                         regions_label=regions_label)
        classification = classifier.classify(force=force)
        if fout:
            cls.output_to_csv(classification, fout)

        return classification



# -*- coding: utf-8 -*-
"""
Convert setback geotiff to .h5 exclusion layers
"""
import logging
import numpy as np
import os
from warnings import warn

from reVX.handlers.geotiff import Geotiff
from reVX.handlers.outputs import Outputs
from reVX.utilities import ExclusionsConverter

logger = logging.getLogger(__name__)


def parse_setbacks(setbacks, chunks=(128, 128), is_inclusion_layer=False):
    """Load setbacks, combine multiple setbacks by state if needed

    This method assumes the data in separate files is non-overlapping.
    In other words, a file containing setbacks exclusions for Illinois
    should not contain any exclusions for Indiana, assuming the setbacks
    for Indiana are in a separate tif file in the same directory.

    Parameters
    ----------
    setbacks : list
        List of paths to setback geotiffs to load and combine
    chunks : tuple, optional
        Chunk size of exclusions in Geotiff, by default (128, 128)
    is_inclusion_layer : bool, optional
        Flag indicating wether this layer should be an inclusion
        layer instead of an exclusion mask, by default False.

    Returns
    -------
    values : ndarray
        Setbacks exclusion array
    """
    logger.info("Merging setbacks...")
    logger.debug("\n\t- ".join([""] + setbacks))
    values = None
    reduction_func = np.minimum if is_inclusion_layer else np.maximum
    for geotiff in setbacks:
        data = ExclusionsConverter.parse_tiff(geotiff, chunks=chunks,
                                              check_tiff=False)[1]
        if values is None:
            values = data
        else:
            values = reduction_func(values, data)

    return values


class SetbacksConverter(ExclusionsConverter):
    """
    Convert setbacks goetiff(s) to excl .h5 layers
    """

    def setbacks_to_layer(self, layer, setbacks, check_tiff=True,
                          is_inclusion_layer=False, transform_atol=0.01,
                          coord_atol=0.001, description=None,
                          scale_factor=None, dtype='uint8'):
        """
        Transfer geotiff setbacks to h5 confirming they match existing layers

        Parameters
        ----------
        layer : str
            Name of layer to create of replace.
        setbacks : str
            Path to geotiff file or directory containing multiple
            geotiff files to be merged.
        check_tiff : bool, optional
            Flag to check tiff profile and coordinates against exclusion
            .h5 profile and coordinates, by default True.
        is_inclusion_layer : bool, optional
            Flag indicating wether this layer should be an inclusion
            layer instead of an exclusion mask, by default False.
        transform_atol : float, optional
            Absolute tolerance parameter when comparing geotiff
            transform data, by default 0.01
        coord_atol : float, optional.
            Absolute tolerance parameter when comparing new un-projected
            geotiff coordinates against previous coordinates,
            by default 0.001.
        description : str, optional
            Description of exclusion layer, by default None.
        scale_factor : int | float, optional
            Scale factor to use to scale geotiff data when added to the
            .h5 file, by default None, which does not apply any scaling.
        dtype : str, optional
            Dtype to save geotiff data as in the .h5 file. Only used
            when 'scale_factor' is not None, by default 'uint8'
        """
        if os.path.isdir(setbacks):
            setbacks = [os.path.join(setbacks, file)
                        for file in os.listdir(setbacks)
                        if file.endswith('.tif')]
        else:
            setbacks = [setbacks]

        logger.debug('\t- Combining setbacks in {}'.format(setbacks))

        if not os.path.exists(self._excl_h5):
            self._init_h5(self._excl_h5, setbacks[0], chunks=self._chunks)

        msg = ("{} is already present in {}".format(layer, self._excl_h5))
        if layer in self.layers:
            if self._replace:
                msg += " and will be replaced"
                logger.warning(msg)
                warn(msg)
            else:
                msg += ", to 'replace' set to True"
                logger.error(msg)
                raise KeyError(msg)

        if check_tiff:
            self._check_geotiff(self._excl_h5, setbacks[0],
                                chunks=self._chunks,
                                transform_atol=transform_atol,
                                coord_atol=coord_atol)

        with Geotiff(setbacks[0], chunks=self._chunks) as tif:
            profile = tif.profile

        setbacks = parse_setbacks(setbacks, chunks=self._chunks,
                                  is_inclusion_layer=is_inclusion_layer)
        if scale_factor is not None:
            setbacks = Outputs._check_data_dtype(setbacks, dtype,
                                                 scale_factor=scale_factor)

        logger.debug('Writing final setback layer to {!r}'
                     .format(self._excl_h5))
        self._write_layer(self._excl_h5, layer, profile, setbacks,
                          chunks=self._chunks, description=description,
                          scale_factor=scale_factor)

    @classmethod
    def layers_to_h5(cls, excl_h5, layers, chunks=(128, 128),
                     replace=True, check_tiff=True,
                     are_inclusion_layers=False, transform_atol=0.01,
                     coord_atol=0.001, descriptions=None, scale_factors=None):
        """
        Create exclusions .h5 file, or load layers into existing exclusion .h5
        file from provided setbacks

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing or to contain exclusion layers
        layers : dict | list
            Dictionary where keys are layer names and values are paths
            to the corresponding geotiff files or paths to directories
            containing multiple geotiff files to be merged for each
            layer. If input is a list of paths to geotiff files, then
            the name of the layer is inferred from the geotiff file
            name.
        chunks : tuple, optional
            Chunk size of exclusions in geotiff, by default (128, 128)
        replace : bool, optional
            Flag to replace existing layers if needed, by default True
        check_tiff : bool, optional
            Flag to check tiff profile and coordinates against exclusion
            .h5 profile and coordinates, by default True.
        are_inclusion_layers : bool, optional
            Flag indicating wether the input layers should be treated
            as inclusion layers instead of exclusion masks,
            by default False.
        transform_atol : float, optional
            Absolute tolerance parameter when comparing geotiff
            transform data, by default 0.01.
        coord_atol : float, optional
            Absolute tolerance parameter when comparing new un-projected
            geotiff coordinates against previous coordinates,
            by default 0.001.
        description : dict, optional
            Description of exclusion layers, by default None
        scale_factor : dict, optional
            Scale factors and dtypes to use when scaling given layers,
            by default None
        """
        if isinstance(layers, list):
            layers = {os.path.basename(lyr).split('.')[0]: lyr
                      for lyr in layers}

        if scale_factors is None:
            scale_factors = {}

        if descriptions is None:
            descriptions = {}

        excls = cls(excl_h5, chunks=chunks, replace=replace)
        logger.info('Creating {}'.format(excl_h5))
        for layer, setbacks in layers.items():
            logger.info('- Transfering {}'.format(layer))
            scale = scale_factors.get(layer, None)
            if scale is not None:
                scale_factor = scale['scale_factor']
                dtype = scale['dtype']
            else:
                scale_factor = None
                dtype = None

            description = descriptions.get(layer, None)

            excls.setbacks_to_layer(layer, setbacks, check_tiff=check_tiff,
                                    is_inclusion_layer=are_inclusion_layers,
                                    transform_atol=transform_atol,
                                    coord_atol=coord_atol,
                                    description=description,
                                    scale_factor=scale_factor,
                                    dtype=dtype)




# -*- coding: utf-8 -*-
"""
Compute setbacks
"""
from .parcel_setbacks import ParcelSetbacks
from .rail_setbacks import RailSetbacks
from .road_setbacks import RoadSetbacks
from .structure_setbacks import StructureSetbacks
from .transmission_setbacks import TransmissionSetbacks
from .water_setbacks import WaterSetbacks

SETBACKS = {'structure': StructureSetbacks,
            'road': RoadSetbacks,
            'rail': RailSetbacks,
            'transmission': TransmissionSetbacks,
            'parcel': ParcelSetbacks,
            'water': WaterSetbacks}



# -*- coding: utf-8 -*-
"""
Compute setbacks exclusions
"""
import logging

from reVX.setbacks.base import AbstractBaseSetbacks, features_clipped_to_county

logger = logging.getLogger(__name__)


class TransmissionSetbacks(AbstractBaseSetbacks):
    """Transmission setbacks.

    These setbacks are computed against a single set of
    transmission features instead of against state level features.
    """

    @staticmethod
    def _feature_filter(features, cnty):
        """Filter the features given a county."""
        return features_clipped_to_county(features, cnty)

    def _regulation_table_mask(self):
        """Return the regulation table mask for setback feature. """
        return self.regulations_table['Feature Type'] == 'transmission'



# -*- coding: utf-8 -*-
"""
Driver class to compute exclusions
"""
import h5py
import json
import os
import logging
from abc import ABC, abstractmethod
from concurrent.futures import as_completed
from warnings import warn

import numpy as np
import geopandas as gpd
import pandas as pd
from pyproj.crs import CRS
import rasterio
from rasterio import features as rio_features
from shapely.geometry import shape

from rex import Outputs
from rex.utilities import SpawnProcessPool, log_mem
from reV.handlers.exclusions import ExclusionLayers
from reVX.handlers.geotiff import Geotiff
from reVX.utilities.utilities import log_versions
from reVX.utilities.exceptions import ExclusionsCheckError

logger = logging.getLogger(__name__)


class AbstractExclusionCalculatorInterface(ABC):
    """Abstract Exclusion Calculator Interface. """

    @property
    @abstractmethod
    def no_exclusions_array(self):
        """np.array: Array representing no exclusions. """
        raise NotImplementedError

    @property
    @abstractmethod
    def exclusion_merge_func(self):
        """callable: Function to merge overlapping exclusion layers. """
        raise NotImplementedError

    @abstractmethod
    def pre_process_regulations(self):
        """Reduce regulations to correct state and features.

        When implementing this method, make sure to update
        `self._regulations.df`.
        """
        raise NotImplementedError

    @abstractmethod
    def parse_features(self):
        """Parse features used to compute exclusions.

        Warnings
        --------
        Use caution when calling this method, especially in multiple
        processes, as the returned feature files may be quite large.
        Reading 100 GB feature files in each of 36 sub-processes will
        quickly overwhelm your RAM.
        """
        raise NotImplementedError

    @abstractmethod
    def _local_exclusions_arguments(self, regulation_value, cnty):
        """Compile and return arguments to `compute_local_exclusions`.

        This method should return a list or tuple of extra args to be
        passed to `compute_local_exclusions`. Do not include the
        `regulation_value` or `cnty`.

        Parameters
        ----------
        regulation_value : float | int
            Regulation value for county.
        cnty : geopandas.GeoDataFrame
            Regulations for a single county.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def compute_local_exclusions(regulation_value, cnty, *args):
        """Compute local feature exclusions.

        This method should compute the exclusions using the information
        about the input county.

        Parameters
        ----------
        regulation_value : float | int
            Regulation value for county.
        cnty : geopandas.GeoDataFrame
            Regulations for a single county.
        *args
            Other arguments required for local exclusion calculation.

        Returns
        -------
        exclusions : np.ndarray
            Array of exclusions.
        """
        raise NotImplementedError

    @abstractmethod
    def compute_generic_exclusions(self, max_workers=None):
        """Compute generic exclusions.

        This method should compute the exclusions using a generic
        regulation value (`self._regulations.generic`).

        Parameters
        ----------
        max_workers : int, optional
            Number of workers to use for exclusions computation, if 1
            run in serial, if > 1 run in parallel with that many
            workers, if `None` run in parallel on all available cores.
            By default `None`.

        Returns
        -------
        exclusions : ndarray
            Raster array of exclusions
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def input_output_filenames(cls, out_dir, features_fpath, kwargs):
        """Generate pairs of input/output file names.

        Parameters
        ----------
        out_dir : str
            Path to output file directory.
        features_fpath : str
            Path to features file. This path can contain
            any pattern that can be used in the glob function.
            For example, `/path/to/features/[A]*` would match
            with all the features in the directory
            `/path/to/features/` that start with "A". This input
            can also be a directory, but that directory must ONLY
            contain feature files. If your feature files are mixed
            with other files or directories, use something like
            `/path/to/features/*.geojson`.
        kwargs : dict
            Dictionary of extra keyword-argument pairs used to
            instantiate the `exclusion_class`.

        Yields
        ------
        tuple
            An input-output filename pair.
        """
        raise NotImplementedError


class AbstractBaseExclusionsMerger(AbstractExclusionCalculatorInterface):
    """
    Create exclusions layers for exclusions
    """

    def __init__(self, excl_fpath, regulations, features, hsds=False):
        """
        Parameters
        ----------
        excl_fpath : str
            Path to .h5 file containing exclusion layers, will also be
            the location of any new exclusion layers
        regulations : `~reVX.utilities.AbstractBaseRegulations` subclass
            A regulations object used to extract exclusion regulation
            values.
        features : str
            Path to file containing features to compute exclusions from.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on
            AWS behind HSDS. By default `False`.
        """
        log_versions(logger)
        self._excl_fpath = excl_fpath
        self._regulations = regulations
        self._features = features
        self._hsds = hsds
        self._fips = self._profile = None
        self._set_profile()
        self._process_regulations(regulations.df)
        self.features = self.parse_features()

    def __repr__(self):
        msg = "{} for {}".format(self.__class__.__name__, self._excl_fpath)
        return msg

    def _set_profile(self):
        """Extract profile from excl h5."""
        with ExclusionLayers(self._excl_fpath, hsds=self._hsds) as f:
            self._profile = f.profile

    def _process_regulations(self, regulations_df):
        """Parse the county regulations.

        Parse regulations, combine with county geometries from
        exclusions .h5 file. The county geometries are intersected with
        features to compute county specific exclusions.

        Parameters
        ----------
        regulations : pandas.DataFrame
            Regulations table

        Returns
        -------
        regulations: `geopandas.GeoDataFrame`
            GeoDataFrame with county level exclusion regulations merged
            with county geometries, use for intersecting with exclusion
            features.
        """
        if regulations_df is None:
            return

        with ExclusionLayers(self._excl_fpath, hsds=self._hsds) as exc:
            self._fips = exc['cnty_fips']
            cnty_fips_profile = exc.get_layer_profile('cnty_fips')

        if 'FIPS' not in regulations_df:
            msg = ('Regulations does not have county FIPS! Please add a '
                   '"FIPS" columns with the unique county FIPS values.')
            logger.error(msg)
            raise RuntimeError(msg)

        if 'geometry' not in regulations_df:
            regulations_df['geometry'] = None

        regulations_df = regulations_df[~regulations_df['FIPS'].isna()]
        regulations_df = regulations_df.set_index('FIPS')

        logger.info('Merging county geometries w/ local regulations')
        shapes_from_raster = rio_features.shapes(
            self._fips.astype(np.int32),
            transform=cnty_fips_profile['transform']
        )
        county_regs = []
        for polygon, fips_code in shapes_from_raster:
            fips_code = int(fips_code)
            if fips_code in regulations_df.index:
                local_regs = regulations_df.loc[[fips_code]].copy()
                local_regs['geometry'] = shape(polygon)
                county_regs.append(local_regs)

        if county_regs:
            regulations_df = pd.concat(county_regs)

        regulations_df = gpd.GeoDataFrame(
            regulations_df,
            crs=self.profile['crs'],
            geometry='geometry'
        )
        regulations_df = regulations_df.reset_index()
        regulations_df = regulations_df.to_crs(crs=self.profile['crs'])
        self._regulations.df = regulations_df

    @property
    def profile(self):
        """dict: Geotiff profile. """
        return self._profile

    @property
    def regulations_table(self):
        """Regulations table.

        Returns
        -------
        geopandas.GeoDataFrame | None
        """
        return self._regulations.df

    @regulations_table.setter
    def regulations_table(self, regulations_table):
        self._process_regulations(regulations_table)

    def _write_exclusions(self, geotiff, exclusions, replace=False):
        """
        Write exclusions to geotiff, replace if requested

        Parameters
        ----------
        geotiff : str
            Path to geotiff file to save exclusions too
        exclusions : ndarray
            Rasterized array of exclusions.
        replace : bool, optional
            Flag to replace local layer data with arr if file already
            exists on disk. By default `False`.
        """
        if os.path.exists(geotiff):
            _error_or_warn(geotiff, replace)

        ExclusionsConverter.write_geotiff(geotiff, self.profile, exclusions)

    def _write_layer(self, out_layer, exclusions, replace=False):
        """Write exclusions to H5, replace if requested

        Parameters
        ----------
        out_layer : str
            Name of new exclusion layer to add to h5.
        exclusions : ndarray
            Rasterized array of exclusions.
        replace : bool, optional
            Flag to replace local layer data with arr if layer already
            exists in the exclusion .h5 file. By default `False`.
        """
        with ExclusionLayers(self._excl_fpath, hsds=self._hsds) as exc:
            layers = exc.layers

        if out_layer in layers:
            _error_or_warn(out_layer, replace)

        try:
            description = self.description
        except AttributeError:
            description = None

        ExclusionsConverter._write_layer(self._excl_fpath, out_layer,
                                         self.profile, exclusions,
                                         description=description)

    def compute_all_local_exclusions(self, max_workers=None):
        """Compute local exclusions for all counties either.

        Parameters
        ----------
        max_workers : int, optional
            Number of workers to use for exclusions computation, if 1
            run in serial, if > 1 run in parallel with that many
            workers, if `None` run in parallel on all available cores.
            By default `None`.

        Returns
        -------
        exclusions : ndarray
            Raster array of exclusions.
        """
        max_workers = max_workers or os.cpu_count()

        log_mem(logger)
        if max_workers > 1:
            logger.info('Computing local exclusions in parallel using {} '
                        'workers'.format(max_workers))
            spp_kwargs = {"max_workers": max_workers,
                          "loggers": [__name__, 'reVX']}
            with SpawnProcessPool(**spp_kwargs) as exe:
                exclusions = self._compute_exclusions_spp(exe, max_workers)
        else:
            logger.info('Computing local exclusions in serial')
            exclusions = None
            for i, (exclusion, cnty) in enumerate(self._regulations):
                args = self._local_exclusions_arguments(exclusion, cnty)
                local_exclusions = self.compute_local_exclusions(exclusion,
                                                                 cnty,
                                                                 *args)
                exclusions = self._combine_exclusions(exclusions,
                                                      local_exclusions,
                                                      cnty['FIPS'].unique())
                logger.debug('Computed exclusions for {} of {} counties'
                             .format((i + 1), len(self.regulations_table)))

        return exclusions

    def _compute_exclusions_spp(self, exe, max_submissions):
        """Compute exclusions in parallel using futures. """
        futures, exclusions = {}, None

        for ind, regulation_info in enumerate(self._regulations, start=1):
            exclusion_value, cnty = regulation_info
            logger.info('Computing exclusions for {}/{} counties'
                        .format(ind, len(self.regulations_table)))
            args = self._local_exclusions_arguments(exclusion_value, cnty)
            future = exe.submit(self.compute_local_exclusions,
                                exclusion_value, cnty, *args)
            futures[future] = cnty['FIPS'].unique()
            if ind % max_submissions == 0:
                exclusions = self._collect_futures(futures, exclusions)
        exclusions = self._collect_futures(futures, exclusions)
        return exclusions

    def _collect_futures(self, futures, exclusions):
        """Collect all futures from the input dictionary. """
        for future in as_completed(futures):
            exclusions = self._combine_exclusions(exclusions,
                                                  future.result(),
                                                  futures.pop(future))
            log_mem(logger)
        return exclusions

    def compute_exclusions(self, out_layer=None, out_tiff=None, replace=False,
                           max_workers=None):
        """
        Compute exclusions for all states either in serial or parallel.
        Existing exclusions are computed if a regulations file was
        supplied during class initialization, otherwise generic exclusions
        are computed.

        Parameters
        ----------
        out_layer : str, optional
            Name to save rasterized exclusions under in .h5 file.
            If `None`, exclusions will not be written to the .h5 file.
            By default `None`.
        out_tiff : str, optional
            Path to save geotiff containing rasterized exclusions.
            If `None`, exclusions will not be written to a geotiff file.
            By default `None`.
        replace : bool, optional
            Flag to replace geotiff if it already exists.
            By default `False`.
        max_workers : int, optional
            Number of workers to use for exclusion computation, if 1 run
            in serial, if > 1 run in parallel with that many workers,
            if `None`, run in parallel on all available cores.
            By default `None`.

        Returns
        -------
        exclusions : ndarray
            Raster array of exclusions
        """
        exclusions = self._compute_merged_exclusions(max_workers=max_workers)

        if out_layer is not None:
            logger.info('Saving exclusion layer to {} as {}'
                        .format(self._excl_fpath, out_layer))
            self._write_layer(out_layer, exclusions, replace=replace)

        if out_tiff is not None:
            logger.debug('Writing exclusions to {}'.format(out_tiff))
            self._write_exclusions(out_tiff, exclusions, replace=replace)

        return exclusions

    def _compute_merged_exclusions(self, max_workers=None):
        """Compute and merge local and generic exclusions, if necessary. """
        mw = max_workers

        if self._regulations.locals_exist:
            self.pre_process_regulations()

        generic_exclusions_exist = self._regulations.generic_exists
        local_exclusions_exist = self._regulations.locals_exist

        if not generic_exclusions_exist and not local_exclusions_exist:
            msg = ("Found no exclusions to compute: No regulations detected, "
                   "and generic multiplier not set.")
            logger.error(msg)
            raise ValueError(msg)

        if generic_exclusions_exist and not local_exclusions_exist:
            return self.compute_generic_exclusions(max_workers=mw)

        if local_exclusions_exist and not generic_exclusions_exist:
            local_excl = self.compute_all_local_exclusions(max_workers=mw)
            # merge ensures local exclusions are clipped county boundaries
            return self._merge_exclusions(None, local_excl)

        generic_exclusions = self.compute_generic_exclusions(max_workers=mw)
        local_exclusions = self.compute_all_local_exclusions(max_workers=mw)
        return self._merge_exclusions(generic_exclusions, local_exclusions)

    def _merge_exclusions(self, generic_exclusions, local_exclusions):
        """Merge local exclusions onto the generic exclusions."""
        logger.info('Merging local exclusions onto the generic exclusions')

        self.pre_process_regulations()
        local_fips = self.regulations_table["FIPS"].unique()
        return self._combine_exclusions(generic_exclusions, local_exclusions,
                                        local_fips, replace_existing=True)

    def _combine_exclusions(self, existing, additional, cnty_fips,
                            replace_existing=False):
        """Combine local exclusions using FIPS code"""
        if existing is None:
            existing = self.no_exclusions_array.astype(additional.dtype)

        local_exclusions = np.isin(self._fips, cnty_fips)
        if replace_existing:
            new_local_exclusions = additional[local_exclusions]
        else:
            new_local_exclusions = self.exclusion_merge_func(
                existing[local_exclusions],
                additional[local_exclusions])

        existing[local_exclusions] = new_local_exclusions
        return existing

    @classmethod
    def run(cls, excl_fpath, features_path, out_dir, regulations,
            max_workers=None, replace=False, out_layers=None, hsds=False,
            **kwargs):
        """
        Compute exclusions and write them to a geotiff. If a regulations
        file is given, compute local exclusions, otherwise compute
        generic exclusions. If both are provided, generic and local
        exclusions are merged such that the local exclusions override
        the generic ones.

        Parameters
        ----------
        excl_fpath : str
            Path to .h5 file containing exclusion layers, will also be
            the location of any new exclusion layers.
        features_path : str
            Path to file or directory feature shape files.
            This path can contain any pattern that can be used in the
            glob function. For example, `/path/to/features/[A]*` would
            match with all the features in the directory
            `/path/to/features/` that start with "A". This input
            can also be a directory, but that directory must ONLY
            contain feature files. If your feature files are mixed
            with other files or directories, use something like
            `/path/to/features/*.geojson`.
        out_dir : str
            Directory to save exclusion geotiff(s) into
        regulations : `~reVX.utilities.AbstractBaseRegulations` subclass
            A regulations object used to extract exclusion regulation
            distances.
        max_workers : int, optional
            Number of workers to use for exclusion computation, if 1 run
            in serial, if > 1 run in parallel with that many workers,
            if `None`, run in parallel on all available cores.
            By default `None`.
        replace : bool, optional
            Flag to replace geotiff if it already exists.
            By default `False`.
        out_layers : dict, optional
            Dictionary mapping feature file names (with extension) to
            names of layers under which exclusions should be saved in
            the `excl_fpath` .h5 file. If `None` or empty dictionary,
            no layers are saved to the h5 file. By default `None`.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on
            AWS behind HSDS. By default `False`.
        **kwargs
            Keyword args to exclusions calculator class.
        """

        out_layers = out_layers or {}
        cls_init_kwargs = {"excl_fpath": excl_fpath,
                           "regulations": regulations}
        cls_init_kwargs.update(kwargs)
        files = cls.input_output_filenames(out_dir, features_path,
                                           cls_init_kwargs)
        for f_in, f_out in files:
            if os.path.exists(f_out) and not replace:
                msg = ('{} already exists, exclusions will not be re-computed '
                       'unless replace=True'.format(f_out))
                logger.error(msg)
            else:
                logger.info("Computing exclusions from {} and saving "
                            "to {}".format(f_in, f_out))
                out_layer = out_layers.get(os.path.basename(f_in))
                exclusions = cls(excl_fpath=excl_fpath,
                                 regulations=regulations, features=f_in,
                                 hsds=hsds, **kwargs)
                exclusions.compute_exclusions(out_tiff=f_out,
                                              out_layer=out_layer,
                                              max_workers=max_workers,
                                              replace=replace)


class ExclusionsConverter:
    """
    Convert exclusion layers between .h5 and .tif (geotiff)
    """
    def __init__(self, excl_h5, hsds=False, chunks=(128, 128), replace=True):
        """
        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing or to contain exclusion layers
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        chunks : tuple, optional
            Chunk size of exclusions in .h5 and Geotiffs, by default (128, 128)
        replace : bool, optional
            Flag to replace existing layers if needed, by default True
        """
        log_versions(logger)
        self._excl_h5 = excl_h5
        self._hsds = hsds
        self._chunks = chunks
        self._replace = replace

    def __repr__(self):
        msg = "{} for {}".format(self.__class__.__name__, self._excl_h5)
        return msg

    def __getitem__(self, layer):
        """
        Parameters
        ----------
        layer : str
            Layer to extract data for

        Returns
        -------
        profile : dict
            Geotiff profile (attributes)
        values : ndarray
            Geotiff data
        """

        if layer not in self.layers:
            msg = "{} is not present in {}".format(layer, self._excl_h5)
            logger.error(msg)
            raise KeyError(msg)

        profile, values = self._extract_layer(self._excl_h5, layer,
                                              hsds=self._hsds)
        return profile, values

    def __setitem__(self, layer, geotiff):
        """
        Parameters
        ----------
        layer : str
            Layer to set
        geotiff : str
            Path to GeoTiff to load data from
        """
        self.geotiff_to_layer(layer, geotiff)

    @property
    def layers(self):
        """
        Available exclusion layers in .h5 file

        Returns
        -------
        layers : list
            Available layers in .h5 file
        """
        with ExclusionLayers(self._excl_h5, hsds=self._hsds) as exc:
            layers = exc.layers

        return layers

    @staticmethod
    def _init_h5(excl_h5, geotiff, chunks=(128, 128)):
        """
        Initialize exclusions .h5 file from geotiff:
        - Transfer profile, shape, and meta

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers
        geotiff : str
            Path to geotiff file
        chunks : tuple
            Chunk size of exclusions in Geotiff
        """
        logger.debug('\t- Initializing {} from {}'
                     .format(excl_h5, geotiff))
        with Geotiff(geotiff, chunks=chunks) as src:
            profile = src.profile
            shape = src.shape
            lat, lon = src.lat_lon
            logger.debug('\t- "profile", "meta", and "shape" extracted from {}'
                         .format(geotiff))

        try:
            with h5py.File(excl_h5, mode='w') as dst:
                dst.attrs['profile'] = json.dumps(profile)
                logger.debug('\t- Default profile:\n{}'.format(profile))
                dst.attrs['shape'] = shape
                logger.debug('\t- Default shape:\n{}'.format(shape))
                dst.attrs['chunks'] = chunks
                logger.debug('\t- Default chunks:\n{}'.format(chunks))

                dst.create_dataset('latitude', shape=lat.shape,
                                   dtype=np.float32, data=lat,
                                   chunks=chunks)
                logger.debug('\t- latitude coordiantes created')

                dst.create_dataset('longitude', shape=lon.shape,
                                   dtype=np.float32, data=lon,
                                   chunks=chunks)
                logger.debug('\t- longitude coordiantes created')
        except Exception:
            logger.exception("Error initilizing {}".format(excl_h5))
            if os.path.exists(excl_h5):
                os.remove(excl_h5)

    @staticmethod
    def _check_crs(baseline_crs, test_crs, ignore_keys=('no_defs',)):
        """
        Compare baseline and test crs values

        Parameters
        ----------
        baseline_crs : dict
            Baseline CRS to use a truth, must be a dict
        test_crs : dict
            Test CRS to compare with baseline, must be a dictionary
        ignore_keys : tuple
            Keys to not check

        Returns
        -------
        bad_crs : bool
            Flag if crs' do not match
        """
        bad_crs = False
        for k, true_v in baseline_crs.items():
            if k not in ignore_keys:
                test_v = test_crs.get(k, true_v)
                if true_v != test_v:
                    bad_crs = True

        return bad_crs

    @classmethod
    def _check_geotiff(cls, excl_h5, geotiff, chunks=(128, 128),
                       transform_atol=0.01, coord_atol=0.001):
        """
        Compare geotiff with exclusion layer, raise any errors

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers
        geotiff : str
            Path to geotiff file
        chunks : tuple
            Chunk size of exclusions in Geotiff
        transform_atol : float
            Absolute tolerance parameter when comparing geotiff transform data.
        coord_atol : float
            Absolute tolerance parameter when comparing new un-projected
            geotiff coordinates against previous coordinates.
        """
        with Geotiff(geotiff, chunks=chunks) as tif:
            with ExclusionLayers(excl_h5) as h5:
                if tif.bands > 1:
                    error = ('{} contains more than one band!'
                             .format(geotiff))
                    logger.error(error)
                    raise ExclusionsCheckError(error)

                if not np.array_equal(h5.shape, tif.shape):
                    error = ('Shape of exclusion data in {} and {} do not '
                             'match!'.format(geotiff, excl_h5))
                    logger.error(error)
                    raise ExclusionsCheckError(error)

                profile = h5.profile
                h5_crs = CRS.from_string(profile['crs']).to_dict()
                tif_crs = CRS.from_string(tif.profile['crs']).to_dict()
                bad_crs = cls._check_crs(h5_crs, tif_crs)
                if bad_crs:
                    error = ('Geospatial "crs" in {} and {} do not match!'
                             '\n {} !=\n {}'
                             .format(geotiff, excl_h5, tif_crs, h5_crs))
                    logger.error(error)
                    raise ExclusionsCheckError(error)

                if not np.allclose(profile['transform'],
                                   tif.profile['transform'],
                                   atol=transform_atol):
                    error = ('Geospatial "transform" in {} and {} do not '
                             'match!\n {} !=\n {}'
                             .format(geotiff, excl_h5, profile['transform'],
                                     tif.profile['transform']))
                    logger.error(error)
                    raise ExclusionsCheckError(error)

                lat, lon = tif.lat_lon
                if not np.allclose(h5.latitude, lat, atol=coord_atol):
                    error = ('Latitude coordinates {} and {} do not match to '
                             'within {} degrees!'
                             .format(geotiff, excl_h5, coord_atol))
                    logger.error(error)
                    raise ExclusionsCheckError(error)

                if not np.allclose(h5.longitude, lon, atol=coord_atol):
                    error = ('Longitude coordinates {} and {} do not match to '
                             'within {} degrees!'
                             .format(geotiff, excl_h5, coord_atol))
                    logger.error(error)
                    raise ExclusionsCheckError(error)

    @classmethod
    def parse_tiff(cls, geotiff, excl_h5=None, chunks=(128, 128),
                   check_tiff=True, transform_atol=0.01, coord_atol=0.001):
        """
        Extract exclusion layer from given geotiff, compare with excl_h5
        if provided

        Parameters
        ----------
        geotiff : str
            Path to geotiff file
        excl_h5 : str, optional
            Path to .h5 file containing exclusion layers, by default None
        chunks : tuple, optional
            Chunk size of exclusions in Geotiff, by default (128, 128)
        check_tiff : bool, optional
            Flag to check tiff profile and coordinates against exclusion .h5
            profile and coordinates, by default True
        transform_atol : float, optional
            Absolute tolerance parameter when comparing geotiff transform data,
            by default 0.01
        coord_atol : float, optional
            Absolute tolerance parameter when comparing new un-projected
            geotiff coordinates against previous coordinates, by default 0.001

        Returns
        -------
        profile : dict
            Geotiff profile (attributes)
        values : ndarray
            Geotiff data
        """
        if excl_h5 is not None and check_tiff:
            cls._check_geotiff(excl_h5, geotiff, chunks=chunks,
                               transform_atol=transform_atol,
                               coord_atol=coord_atol)

        with Geotiff(geotiff, chunks=chunks) as tif:
            profile, values = tif.profile, tif.values

        return profile, values

    @staticmethod
    def _write_layer(excl_h5, layer, profile, values, chunks=(128, 128),
                     description=None, scale_factor=None):
        """
        Write exclusion layer to .h5 file

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers
        layer : str
            Dataset name in .h5 file
        profile : dict
            Geotiff profile (attributes)
        values : ndarray
            Geotiff data
        chunks : tuple
            Chunk size of dataset in .h5 file
        description : str
            Description of exclusion layer
        scale_factor : int | float, optional
            Scale factor to use to scale geotiff data when added to the .h5
            file, by default None
        """
        if len(chunks) < 3:
            chunks = (1, ) + chunks

        if values.ndim < 3:
            values = np.expand_dims(values, 0)

        with h5py.File(excl_h5, mode='a') as f:
            if layer in f:
                ds = f[layer]
                ds[...] = values
                logger.debug('\t- {} values replaced'.format(layer))
            else:
                ds = f.create_dataset(layer, shape=values.shape,
                                      dtype=values.dtype, chunks=chunks,
                                      data=values)
                logger.debug('\t- {} created and loaded'.format(layer))

            ds.attrs['profile'] = json.dumps(profile)
            logger.debug('\t- Unique profile for {} added:\n{}'
                         .format(layer, profile))
            if description is not None:
                ds.attrs['description'] = description
                logger.debug('\t- Description for {} added:\n{}'
                             .format(layer, description))

            if scale_factor is not None:
                ds.attrs['scale_factor'] = scale_factor
                logger.debug('\t- scale_factor for {} added:\n{}'
                             .format(layer, scale_factor))

    @classmethod
    def _geotiff_to_h5(cls, excl_h5, layer, geotiff, chunks=(128, 128),
                       check_tiff=True, transform_atol=0.01, coord_atol=0.001,
                       description=None, scale_factor=None, dtype='int16'):
        """
        Transfer geotiff exclusions to h5 confirming they match existing layers

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers
        layer : str
            Layer to extract
        geotiff : str
            Path to geotiff file
        chunks : tuple, optional
            Chunk size of exclusions in Geotiff, by default (128, 128)
        check_tiff : bool, optional
            Flag to check tiff profile and coordinates against exclusion .h5
            profile and coordinates, by default True
        transform_atol : float, optional
            Absolute tolerance parameter when comparing geotiff transform data,
            by default 0.01
        coord_atol : float, optional
            Absolute tolerance parameter when comparing new un-projected
            geotiff coordinates against previous coordinates, by default 0.001
        description : str, optional
            Description of exclusion layer, by default None
        scale_factor : int | float, optional
            Scale factor to use to scale geotiff data when added to the .h5
            file, by default None
        dtype : str, optional
            Dtype to save geotiff data as in the .h5 file. Only used when
            'scale_factor' is not None, by default 'int16'
        """
        logger.debug('\t- {} being extracted from {} and added to {}'
                     .format(layer, geotiff, os.path.basename(excl_h5)))

        profile, values = cls.parse_tiff(
            geotiff, excl_h5=excl_h5, chunks=chunks, check_tiff=check_tiff,
            transform_atol=transform_atol, coord_atol=coord_atol)

        if scale_factor is not None:
            attrs = {'scale_factor': scale_factor}
            values = Outputs._check_data_dtype(layer, values, dtype,
                                               attrs=attrs)

        cls._write_layer(excl_h5, layer, profile, values,
                         chunks=chunks, description=description,
                         scale_factor=scale_factor)

    @staticmethod
    def write_geotiff(geotiff, profile, values):
        """
        Write values to geotiff with given profile

        Parameters
        ----------
        geotiff : str
            Path to geotiff file to save data to
        profile : dict
            Geotiff profile (attributes)
        values : ndarray
            Geotiff data
        """
        out_dir = os.path.dirname(geotiff)
        if not os.path.exists(out_dir):
            logger.debug("Creating {}".format(out_dir))
            os.makedirs(out_dir)

        if values.shape[0] != 1:
            values = np.expand_dims(values, 0)

        dtype = values.dtype.name
        profile['dtype'] = dtype
        if np.issubdtype(dtype, np.integer):
            dtype_max = np.iinfo(dtype).max
        else:
            dtype_max = np.finfo(dtype).max

        profile['nodata'] = dtype_max

        with rasterio.open(geotiff, 'w', **profile) as f:
            f.write(values)
            logger.debug('\t- {} created'.format(geotiff))

    @classmethod
    def _extract_layer(cls, excl_h5, layer, geotiff=None, hsds=False):
        """
        Extract given layer from exclusions .h5 file and write to geotiff .tif

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers
        layer : str
            Layer to extract
        geotiff : str
            Path to geotiff file
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS

        Returns
        -------
        profile : dict
            Geotiff profile (attributes)
        values : ndarray
            Geotiff data
        """
        logger.debug('\t - Extracting {} from {}'
                     .format(layer, os.path.basename(excl_h5)))
        with ExclusionLayers(excl_h5, hsds=hsds) as f:
            profile = f.get_layer_profile(layer)
            values = f.get_layer_values(layer)

        if geotiff is not None:
            logger.debug('\t- Writing {} to {}'.format(layer, geotiff))
            cls.write_geotiff(geotiff, profile, values)

        return profile, values

    def geotiff_to_layer(self, layer, geotiff, check_tiff=True,
                         transform_atol=0.01, coord_atol=0.001,
                         description=None, scale_factor=None, dtype='int16'):
        """
        Transfer geotiff exclusions to h5 confirming they match existing layers

        Parameters
        ----------
        layer : str
            Layer to extract
        geotiff : str
            Path to geotiff file
        check_tiff : bool, optional
            Flag to check tiff profile and coordinates against exclusion .h5
            profile and coordinates, by default True
        transform_atol : float, optional
            Absolute tolerance parameter when comparing geotiff transform data,
            by default 0.01
        coord_atol : float, optional
            Absolute tolerance parameter when comparing new un-projected
            geotiff coordinates against previous coordinates, by default 0.001
        description : str, optional
            Description of exclusion layer, by default None
        scale_factor : int | float, optional
            Scale factor to use to scale geotiff data when added to the .h5
            file, by default None
        dtype : str, optional
            Dtype to save geotiff data as in the .h5 file. Only used when
            'scale_factor' is not None, by default 'int16'
        """
        if not os.path.exists(self._excl_h5):
            self._init_h5(self._excl_h5, geotiff, chunks=self._chunks)

        if layer in self.layers:
            msg = ("{} is already present in {}"
                   .format(layer, self._excl_h5))
            if self._replace:
                msg += " and will be replaced"
                logger.warning(msg)
                warn(msg)
            else:
                msg += ", to 'replace' set to True"
                logger.error(msg)
                raise KeyError(msg)

        self._geotiff_to_h5(self._excl_h5, layer, geotiff,
                            chunks=self._chunks,
                            check_tiff=check_tiff,
                            transform_atol=transform_atol,
                            coord_atol=coord_atol,
                            description=description,
                            scale_factor=scale_factor,
                            dtype=dtype)

    def layer_to_geotiff(self, layer, geotiff):
        """
        Extract desired layer from .h5 file and write to geotiff .tif

        Parameters
        ----------
        layer : str
            Layer to extract
        geotiff : str
            Path to geotiff file
        """
        self._extract_layer(self._excl_h5, layer, geotiff=geotiff,
                            hsds=self._hsds)

    @classmethod
    def layers_to_h5(cls, excl_h5, layers, chunks=(128, 128),
                     replace=True, check_tiff=True,
                     transform_atol=0.01, coord_atol=0.001,
                     descriptions=None, scale_factors=None):
        """
        Create exclusions .h5 file, or load layers into existing exclusion .h5
        file from provided geotiffs

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing or to contain exclusion layers
        layers : list | dict
            List of geotiffs to load
            or dictionary mapping goetiffs to the layers to load
        chunks : tuple, optional
            Chunk size of exclusions in Geotiff, by default (128, 128)
        replace : bool, optional
            Flag to replace existing layers if needed, by default True
        check_tiff : bool, optional
            Flag to check tiff profile and coordinates against exclusion .h5
            profile and coordinates, by default True
        transform_atol : float, optional
            Absolute tolerance parameter when comparing geotiff transform data,
            by default 0.01
        coord_atol : float, optional
            Absolute tolerance parameter when comparing new un-projected
            geotiff coordinates against previous coordinates, by default 0.001
        description : dict, optional
            Description of exclusion layers, by default None
        scale_factor : dict, optional
            Scale factors and dtypes to use when scaling given layers,
            by default None
        """
        if isinstance(layers, list):
            layers = {os.path.basename(lyr).split('.')[0]: lyr
                      for lyr in layers}

        if descriptions is None:
            descriptions = {}

        if scale_factors is None:
            scale_factors = {}

        excls = cls(excl_h5, chunks=chunks, replace=replace)
        logger.info('Creating {}'.format(excl_h5))
        for layer, geotiff in layers.items():
            logger.info('- Transfering {}'.format(layer))
            description = descriptions.get(layer, None)
            scale = scale_factors.get(layer, None)
            if scale is not None:
                scale_factor = scale['scale_factor']
                dtype = scale['dtype']
            else:
                scale_factor = None
                dtype = None

            excls.geotiff_to_layer(layer, geotiff, check_tiff=check_tiff,
                                   transform_atol=transform_atol,
                                   coord_atol=coord_atol,
                                   description=description,
                                   scale_factor=scale_factor,
                                   dtype=dtype)

    @classmethod
    def extract_layers(cls, excl_h5, layers, chunks=(128, 128),
                       hsds=False):
        """
        Extract given layers from exclusions .h5 file and save to disk
        as GeoTiffs

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing or to contain exclusion layers
        layers : dict
            Dictionary mapping layers to geotiffs to create
        chunks : tuple
            Chunk size of exclusions in .h5 and Geotiffs
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        """
        excls = cls(excl_h5, chunks=chunks, hsds=hsds)
        logger.info('Extracting layers from {}'.format(excl_h5))
        for layer, geotiff in layers.items():
            logger.info('- Extracting {}'.format(geotiff))
            excls.layer_to_geotiff(layer, geotiff)

    @classmethod
    def extract_all_layers(cls, excl_h5, out_dir, chunks=(128, 128),
                           hsds=False):
        """
        Extract all layers from exclusions .h5 file and save to disk
        as GeoTiffs

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing or to contain exclusion layers
        out_dir : str
            Path to output directory into which layers should be saved as
            GeoTiffs
        chunks : tuple
            Chunk size of exclusions in .h5 and Geotiffs
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        """
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)

        excls = cls(excl_h5, chunks=chunks, hsds=hsds)
        logger.info('Extracting layers from {}'.format(excl_h5))
        for layer in excls.layers:
            geotiff = os.path.join(out_dir, "{}.tif".format(layer))
            logger.info('- Extracting {}'.format(geotiff))
            excls.layer_to_geotiff(layer, geotiff)


def _error_or_warn(name, replace):
    """If replace, throw warning, otherwise throw error. """
    if not replace:
        msg = ('{} already exists. To replace it set "replace=True"'
               .format(name))
        logger.error(msg)
        raise IOError(msg)

    msg = ('{} already exists and will be replaced!'.format(name))
    logger.warning(msg)
    warn(msg)
